/* scl2wav.cpp
This program converts NSWCCD Code 653's scl 32-bit binary data format to 32-bit float wav file format.
This is useful for loading, analyzing, or compressing test data using audio programs.
The SCL code was copied from JHSV Analysis Engine.

Gregory Alan Hildstrom
Hildstrom Engineering

version: 200708021515
*/

#include<iostream>//for cout and standard i/o
#include<fstream>//for file i/o
#include<vector>//for STL vector
#include<string>//for string
using namespace std;//using the standard namespace


//global variables - single values
static int numchan = 0;//number of real channels in input data file
static float scanrate = 0;//scan rate of input data file
static double cal = 0;//temporary cal factor from input data file
static float datapt = 0;//temporary data point from input data file
//global variables - time history storage
static vector<double>cals;//vector stores all cal factors from input data file
static vector< vector<float> > data;//vector stores actual time histories of input data file


//for conversion between high-byte-first and low-byte-first of 32-bit integers
static int swapbytes(int input){
	int output = 0;
	((char*)&output)[0] = ((char*)&input)[3];
	((char*)&output)[3] = ((char*)&input)[0];
	((char*)&output)[1] = ((char*)&input)[2];
	((char*)&output)[2] = ((char*)&input)[1];
	return output;
}//end swapbytes


//for conversion between high-byte-first and low-byte-first of 32-bit floating point number
static float swapbytes(float input){
	float output = 0;
	((char*)&output)[0] = ((char*)&input)[3];
	((char*)&output)[3] = ((char*)&input)[0];
	((char*)&output)[1] = ((char*)&input)[2];
	((char*)&output)[2] = ((char*)&input)[1];
	return output;
}//end swapbytes


//for conversion between high-byte-first and low-byte-first 64-bit floating point numbers
static double swapbytes(double input){
	double output = 0;
	((char*)&output)[0] = ((char*)&input)[7];
	((char*)&output)[7] = ((char*)&input)[0];
	((char*)&output)[1] = ((char*)&input)[6];
	((char*)&output)[6] = ((char*)&input)[1];
	((char*)&output)[2] = ((char*)&input)[5];
	((char*)&output)[5] = ((char*)&input)[2];
	((char*)&output)[3] = ((char*)&input)[4];
	((char*)&output)[4] = ((char*)&input)[3];
	return output;
}//end swapbytes


//This function loads a standard SCL binary data file
//as defined in "JHSV Analysis Engine" - NSWCCD-65-TR-2006/15.
static void loadSCLfile(string filename){
	ifstream inputfile;//the input filestream
	inputfile.open(filename.data(), ios::binary);//open the desired file in binary mode

	if(!inputfile.good()){//if the input file cannot be read
		cout << "Error: Cannot read input file " << filename << ".\n";
		cerr << "Error: Cannot read input file " << filename << ".\n";
		inputfile.close();
		exit(1);//exit completely with an error
	}//end if

	//read number of channels
	inputfile.read((char*)&numchan,4);//read the 4-byte integer
	numchan = swapbytes(numchan);//swap bytes from HBF to LBF
	if(numchan > 10000){//probable error
		cout << "Error: Numchan " << numchan << " > 10000 channels.\n";
		cerr << "Error: Numchan " << numchan << " > 10000 channels.\n";
		inputfile.close();
		exit(1);//exit completely with an error
	}//end if
	cout << "numchan: " << numchan << endl;

	//read scanrate (model/recorded scale)
	inputfile.read((char*)&scanrate,4);//read the 4-byte float
	scanrate = swapbytes(scanrate);//swap bytes from HBF to LBF
	cout << "scanrate: " << scanrate << endl;

	//read all cal factors
	cals.resize(numchan);//resize the cals vector/array to 1 for each channel
	for(int i=0; i<numchan; i++){
		inputfile.read((char*)&cal,8);//read the 8-byte doubles
		cal = swapbytes(cal);//swap bytes from HBF to LBF
		cals[i] = cal;//store 1 cal for each channel in cals vector
	}//end for

	//ready to read interleaved data at this point using numchan until end of file
	data.resize(numchan);//resize the data vector/array of arrays to 1 vector/array for each channel
	while(!inputfile.eof()){//continue until the end of file is reached
		for(int i=0; i<numchan && !inputfile.eof(); i++){//for each channel
			inputfile.read((char*)&datapt,4);//read the 4-byte float
			if(!inputfile.eof()){//if we are not at the end of the file
				datapt = swapbytes(datapt);//swap bytes from HBF to LBF
				data[i].insert(data[i].end(),datapt);//save the data point at the end of this channel's vector/array
			}//end if
		}//end for
	}//end while

	inputfile.close();//close the input file stream
}//end loadSCLfile


//This function outputs a standard SCL binary data file
static void saveSCLfile(string filename){

	unsigned int i = 0;
	unsigned int j = 0;

	//generate new file name with .scl
	string fname = filename;
	fname[fname.size()-3] = 's';
	fname[fname.size()-2] = 'c';
	fname[fname.size()-1] = 'l';

	//open and overwrite any existing binary output file
	ofstream outputfile;
	outputfile.open(fname.data(), ios::trunc|ios::binary);

	//scl header number of channels
	int tempnumchan = swapbytes(numchan);
	outputfile.write((char*)&tempnumchan,4);

	//scl header scan rate
	float tempscanrate = swapbytes(scanrate);
	outputfile.write((char*)&tempscanrate,4);

	//scl header cal factors are the max value in the histogram for each channel
	double tempcal;
	for(i=0; i < (unsigned int)numchan; i++){
		tempcal = swapbytes(cals[i]);
		outputfile.write((char*)&tempcal,8);
	}//end for

	//scl interleaved data
	for(i=0; i < data[0].size(); i++){
		for(j=0; j < (unsigned int)numchan; j++){
			datapt = swapbytes(data[j][i]);
			outputfile.write((char*)&datapt, 4);
		}//end for
	}//end for

	outputfile.close();

}//end saveSCLfile

/*
Offset   Length   Contents
  0      4 bytes  "RIFF"
  4      4 bytes  <File length - 8>
  8      4 bytes  "WAVE"

 12      4 bytes  "fmt "
 16      4 bytes  <Length of the fmt data> // (=16)
 20      2 bytes  <WAVE File Encoding Tag>
 22      2 bytes  <Channels>     // Channels: 1 = mono, 2 = stereo
 24      4 bytes  <Sample rate>  // Samples per second: e.g., 44100
 28      4 bytes  <bytes/second> // sample rate * block align
 32      2 bytes  <block align>  // channels * bits/sample / 8
 34      2 bytes  <bits/sample>  // 8 or 16

 36      4 bytes  "data"
 40      4 bytes  <Sample data size(n)>
 44     (n)bytes  <Sample data>
 */
static void loadWAVfile(string filename){
	ifstream input;
	input.open(filename.data());
	if(!input.good()){
		cerr << "Error: could not open input file: " << filename << endl;
		exit(1);
	}//end if
	
	char sample[5];
	sample[4] = '\0';
	char tag[5];
	tag[4] = '\0';
	
	input.read(sample,4);
	if((string)sample != (string)"RIFF"){
		cerr << "Error: not a RIFF file: " << sample << endl;
		exit(1);
	}//end if
	cout << "riff: " << sample << endl;
	
	unsigned int filesize = 0;
	input.read((char*)&filesize,4);
	if(filesize == 0){
		cerr << "Error: invalid file size: " << filesize << endl;
		exit(1);
	}//end if
	cout << "filesize: " << filesize << endl;
	
	input.read(sample,4);
	if((string)sample != "WAVE"){
		cerr << "Error: not a WAVE file: " << sample << endl;
		exit(1);
	}//end if
	cout << "wave: " << sample << endl;
	
	input.read(sample,4);
	if((string)sample != "fmt "){
		cerr << "Error: no format header: " << sample << endl;
		exit(1);
	}//end if
	cout << "fmt: " << sample << endl;
	
	int fmtsize = 0;
	input.read((char*)&fmtsize,4);
	if(fmtsize != 16){
		cerr << "Error: invalid fmtsize: " << fmtsize << endl;
		exit(1);
	}//end if
	cout << "fmtsize: " << fmtsize << endl;
	
	short encoding = 0;
	input.read((char*)&encoding,2);
	if(encoding != 3){
		cerr << "Error: invalid encoding: " << encoding << endl;
		exit(1);
	}//end if
	cout << "encoding: " << encoding << endl;
	
	short tempnumchan = 0;
	input.read((char*)&tempnumchan,2);
	numchan = tempnumchan;
	if(numchan <= 0){
		cerr << "Error: numchan: " << numchan << endl;
		exit(1);
	}//end if
	data.resize(numchan);
	cals.resize(numchan);
	cout << "numchan: " << numchan << endl;
	
	int samplerate = 0;
	input.read((char*)&samplerate,4);
	scanrate = samplerate;
	if(scanrate <= 0){
		cerr << "Error: scan rate: " << scanrate << endl;
		exit(1);
	}//end if
	cout << "scanrate: " << scanrate << endl;
	
	int bytespersecond = 0;
	input.read((char*)&bytespersecond,4);
	if(bytespersecond <= 0){
		cerr << "Error: bytespersecond: " << bytespersecond << endl;
		exit(1);
	}//end if
	cout << "bytespersecond: " << bytespersecond << endl;
	
	short blockalign = 0;
	input.read((char*)&blockalign,2);
	if(blockalign <= 0){
		cerr << "Error: blockalign: " << blockalign << endl;
		exit(1);
	}//end if
	cout << "blockalign: " << blockalign << endl;
	
	short bitspersample = 0;
	input.read((char*)&bitspersample,2);
	if(bitspersample != 32){
		cerr << "Error: bitspersample: " << bitspersample << endl;
		exit(1);
	}//end if
	cout << "bitspersample: " << bitspersample << endl;
	
	int chunksize = 0;
	while(input.good()){
		input.read(tag,4);
		if(input.good()){
			cout << "tag: " << tag << endl;
			input.read((char*)&chunksize,4);
			cout << "chunksize: " << chunksize << endl;
		}//end if
		if((string)tag != (string)"data" && (string)tag != (string)"cals" && input.good()){
			cerr << "Error: unrecognized tag: " << tag << " at " << input.tellg() << endl;
			for(int i=0; i < chunksize; i++) input.read(sample,1);
		}//end if
		for(int i=0; i < numchan && input.good() && (string)tag == (string)"cals"; i++){
			input.read((char*)&cal,8);
			cals.insert(cals.end(),cal);
		}//end for
		for(int i=0; i < chunksize/4/numchan && input.good() && (string)tag == (string)"data"; i++){
			for(int j=0; j < numchan; j++){
				input.read((char*)&datapt,4);
				data[j].insert(data[j].end(),datapt);
			}//end for
		}//end for
	}//end while
	
	input.close();
	
}//end loadWAVfile


/*
Offset   Length   Contents
  0      4 bytes  "RIFF"
  4      4 bytes  <File length - 8>
  8      4 bytes  "WAVE"

 12      4 bytes  "fmt "
 16      4 bytes  <Length of the fmt data> // (=16)
 20      2 bytes  <WAVE File Encoding Tag>
 22      2 bytes  <Channels>     // Channels: 1 = mono, 2 = stereo
 24      4 bytes  <Sample rate>  // Samples per second: e.g., 44100
 28      4 bytes  <bytes/second> // sample rate * block align
 32      2 bytes  <block align>  // channels * bits/sample / 8
 34      2 bytes  <bits/sample>  // 8 or 16

 36      4 bytes  "data"
 40      4 bytes  <Sample data size(n)>
 44     (n)bytes  <Sample data>
 */
static void saveWAVfile(string filename){
	ofstream output;
	output.open(filename.data(), ios::trunc|ios::binary);
	if(!output.good()){
		cerr << "Error: could not open output file: " << filename << endl;
		exit(1);
	}//end if
	
	char sample[4];
	char tag[4];
	
	output.write("RIFF",4);
	
	unsigned int filesize = 4+4+4+16+(4+4+numchan*8)+data[0].size()*(4+4+numchan*4);
	output.write((char*)&filesize,4);
	
	output.write("WAVE",4);
	
	output.write("fmt ",4);
	
	int fmtsize = 16;
	output.write((char*)&fmtsize,4);
	
	short encoding = 3;
	output.write((char*)&encoding,2);
	
	short tempnumchan = numchan;
	output.write((char*)&tempnumchan,2);
	
	int samplerate = 0;
	samplerate = (int)scanrate;
	output.write((char*)&samplerate,4);
	
	int bytespersecond = samplerate * numchan * 32 / 8;
	output.write((char*)&bytespersecond,4);
	
	short blockalign = numchan * 32 / 8;
	output.write((char*)&blockalign,2);
	
	short bitspersample = 32;
	output.write((char*)&bitspersample,2);

	int chunksize = numchan*8;
	output.write("cals",4);
	output.write((char*)&chunksize,4);
	for(int i=0; i < numchan && output.good(); i++){
		output.write((char*)&cals[i],8);
	}//end for
	
	chunksize = numchan*4*data[0].size();
	output.write("data",4);
	output.write((char*)&chunksize,4);
	for(int j=0; j < data[0].size(); j++){
		for(int i=0; i < numchan && output.good(); i++){
			output.write((char*)&data[i][j],4);
		}//end for
	}//end for
	
	output.close();
	
}//end saveWAVfile


int main(int argc, char** argv){
	if(argc != 2 && argc != 3){
		cerr << "Error: " << argv[0] << " requires 1 or two arguments\n";
		cerr << "Usages:\n";
		cerr << argv[0] << " input.scl\n";
		cerr << argv[0] << " input.wav\n";
		cerr << argv[0] << " input.scl output.wav\n";
		cerr << argv[0] << " input.wav output.scl\n";
		exit(1);
	}//end if
	
	string infile = argv[1];
	string outfile;
	if(argc == 2){
		outfile = infile;
		if(infile[infile.size()-3] == 'w' && infile[infile.size()-2] == 'a' && infile[infile.size()-1] == 'v'){
			outfile[outfile.size()-3] = 's';
			outfile[outfile.size()-2] = 'c';
			outfile[outfile.size()-1] = 'l';
		}//end if
		else if(infile[infile.size()-3] == 's' && infile[infile.size()-2] == 'c' && infile[infile.size()-1] == 'l'){
			outfile[outfile.size()-3] = 'w';
			outfile[outfile.size()-2] = 'a';
			outfile[outfile.size()-1] = 'v';
		}//end if
		else{
			cerr << "Error: 1 only .scl and .wav files are supported: " << infile << " " << outfile << endl;
			exit(1);
		}//end if
	}//end if	
	else outfile = argv[2];
		
	if(infile[infile.size()-3] == 's' && infile[infile.size()-2] == 'c' && infile[infile.size()-1] == 'l' &&
	outfile[outfile.size()-3] == 'w' && outfile[outfile.size()-2] == 'a' && outfile[outfile.size()-1] == 'v'){
		loadSCLfile(infile);
		saveWAVfile(outfile);
	}//end if
	else if(infile[infile.size()-3] == 'w' && infile[infile.size()-2] == 'a' && infile[infile.size()-1] == 'v' &&
	outfile[outfile.size()-3] == 's' && outfile[outfile.size()-2] == 'c' && outfile[outfile.size()-1] == 'l'){
		loadWAVfile(infile);
		saveSCLfile(outfile);
	}//end if
	else{
		cerr << "Error: 2 only .scl and .wav files are supported: " << infile << " " << outfile << endl;
		exit(1);
	}//end if	
	
	return 0;
}//end main

